/* 
+-----------+
| permute   |		
+-----------+


permutations on a list numbers;
in lexicographical order [Dijkstra algoritm]

<m@ash.to> 

*/

#include "m_pd.h"
#include <stdlib.h>



static t_class *permute_class;

typedef struct permute
{
    t_object x_obj;   
	t_outlet *f_out, *b_out;
	
    t_int *value;
	t_int N;
	
	t_int count;
	t_int max;
	
	char listmode;
    
} t_permute;



void getNext(t_permute *x)
{

	// permutation g in the lexicographic (alphabetic) (the Dijkstra algorithm.)
	// found through: http://www.cut-the-knot.com/do_you_know/AllPerm.shtml 
	// && addionally at: http://mandala.co.uk/permutations/
	// should be described at E.W.Dijkstra, A Discipline of Programming, Prentice-Hall, 1976, p71


	t_int N = x->N;

	t_int j = N;
  	t_int i = N - 1;
	t_int t;

  	while(x->value[i-1] >= x->value[i]) i = i-1;
		
	while (x->value[j-1] <= x->value[i-1]) j = j-1;
	
	t = x->value[i-1];
	x->value[i-1] = x->value[j-1];    // swap values at positions (i-1) and (j-1)
	x->value[j-1] = t;

	i++; 
	j = N;

	while (i < j)
	{
		t = x->value[i-1];
		x->value[i-1] = x->value[j-1];    // swap values at positions (i-1) and (j-1)
		x->value[j-1] = t;			    
		
		i++;		
		j--;
	}	
}

t_int fak(t_int n)
{
  if (n<=1)
    return 1;
  return n*fak(n-1);
}


void permute_ouput_values(t_permute *x)
{
	t_int i;
	
	t_atom list[x->N];
	
	//single out	
	// for(i=0;i<x->N;i++) outlet_float(x->f_out, x->value[i]);
	
	if(x->listmode==0) 	// list with 2 values: index and value
	{		
		for(i=0;i<x->N;i++) 
		{
			SETFLOAT(list,i);
			SETFLOAT(list+1, x->value[i]);			
			outlet_list(x->f_out, gensym("list"), 2, list);
		}
	}
	else if (x->listmode==1) 	// as long list
	{
		for(i=0;i<x->N;i++) 
			SETFLOAT(list + i,x->value[i]);
		outlet_list(x->f_out, NULL, x->N, list);
	}		
}


void permute_set(t_permute *x, t_floatarg f)
{
	t_int i;
	
	free(x->value);
	x->N = f; 
	x->value = malloc(x->N * sizeof(t_int));
	x->count = 0;
	x->max = fak(x->N);

	for(i=0;i<x->N;i++) x->value[i] = i;	
}

void permute_toggle_listmode(t_permute *x, t_floatarg f)
{
	if (f==0) x->listmode = 0;
	else if (f==1) x->listmode = 1;	
}

void permute_reset(t_permute *x)
{	
	t_int i;
	
	x->count = 0;	
	for(i=0;i<x->N;i++) x->value[i] = i;	
}


void permute_bang(t_permute *x)
{
	if (x->count != 0) getNext(x);
	permute_ouput_values(x);
	x->count++;	

	if(x->count == x->max) 
	{
		outlet_bang(x->b_out);
		permute_reset(x);
	}	
}

void permute_cycle(t_permute *x)
{
	t_int i,t;
		
	t = x->value[0];
	for(i=0; i<x->N-1; i++) x->value[i] = x->value[i+1];	
	x->value[x->N-1] = t;
	
	permute_ouput_values(x);
}


void permute_free(t_permute *x)
{
  	free(x->value);
}

void *permute_new(t_floatarg f)
{
    t_permute *x = (t_permute *)pd_new(permute_class);		       
	x->f_out = outlet_new(&x->x_obj, &s_list);
	x->b_out = outlet_new(&x->x_obj, &s_bang);	

	 x->listmode = 0;
	permute_set(x,f);	

    return (x);
}


void permute_setup(void)
{    
    permute_class = class_new(gensym("permute"), (t_newmethod)permute_new, 
		(t_method)permute_free, sizeof(t_permute), 0, A_DEFFLOAT, 0);

    class_addbang(permute_class, permute_bang);
    class_addmethod(permute_class, (t_method)permute_reset, gensym("reset"), 0);  
	class_addmethod(permute_class, (t_method)permute_set, gensym("set"),A_DEFFLOAT, 0);
	class_addmethod(permute_class, (t_method)permute_toggle_listmode, gensym("listmode"),A_DEFFLOAT, 0);	
	class_addmethod(permute_class, (t_method)permute_cycle, gensym("cycle"), 0);	
	
	class_sethelpsymbol(permute_class, gensym("help-permute"));

}
